from typing import List, Dict
from MLVisualizationTools.backend import colinfo
import pandas as pd

#Functions for passing data to ml models

#region Tensorflow
#region grid
def TFModelPredictionGrid(model, x:str, y:str, data:pd.DataFrame,
                          exclude:List[str] = None, steps:int=20) -> pd.DataFrame:
    """
    Creates a dataset from a 2d prediction on a tensorflow model. Wrapper function for TFModelPredictionGridRaw()
    that automatically handles column info generation.

    :param model: A tensorflow model
    :param x: xaxis for graph data
    :param y: yaxis for graph data
    :param data: A pandas dataframe
    :param exclude: Values to be excluded from data, useful for output values
    :param steps: Resolution to scan model with
    """
    return TFModelPredictionGridRaw(model, x, y, colinfo(data, exclude), steps)

def TFModelPredictionGridRaw(model, x:str, y:str, coldata:List[Dict], steps:int=20) -> pd.DataFrame:
    """
    Creates a dataset from a 2d prediction on a tensorflow model. Wrapper function for TFModelPredictionGridRaw()
    that automatically handles column info generation.

    Call from TFModelPredictionGrid to autogen params.

    Coldata should be formatted with keys 'name', 'min', 'max', 'mean'

    :param model: A tensorflow model
    :param model: A tensorflow model
    :param x: xaxis for graph data
    :param y: yaxis for graph data
    :param coldata: An ordered list of dicts with col names, min max values, and means
    :param steps: Resolution to scan model with
    """
    allcols = []
    for item in coldata:
        allcols.append(item['name'])

    assert x in allcols, "X must be in coldata"
    assert y in allcols, "Y must be in coldata"

    cols = []
    for item in coldata:
        if item not in [x, y]:
            cols.append(item['name'])

    srow = []
    for item in cols:
        for d in coldata:
            if d['name'] == item:
                srow.append(d['mean'])

    srow = [srow] * (steps ** 2)
    preddata = pd.DataFrame(srow, columns=cols)

    col = []
    for pos in range(0, steps):
        for item in coldata:
            if item['name'] == x:
                col.append(pos * (item['max'] - item['min']) / (steps - 1) + item['min'])
    col = col * steps
    preddata[x] = col

    col = []
    for pos in range(0, steps):
        for item in coldata:
            if item['name'] == y:
                col += [pos * (item['max'] - item['min']) / (steps - 1) + item['min']] * steps
    preddata[y] = col

    predictions = model.predict(preddata)
    preddata['Output'] = predictions
    return preddata
#endregion grid

#region animation
def TFModelPredictionAnimation(model, x:str, y:str, anim:str, data: pd.DataFrame,
                               exclude:List[str] = None, steps:int=20) -> pd.DataFrame:
    """
    Creates a dataset from a 2d prediction on a tensorflow model. Wrapper function for TFModelPredictionGridRaw()
    that automatically handles column info generation.

    :param model: A tensorflow model
    :param x: xaxis for graph data
    :param y: yaxis for graph data
    :param anim: Animation axis for graph data
    :param data: A pandas dataframe
    :param exclude: Values to be excluded from data, useful for output values
    :param steps: Resolution to scan model with
    """
    return TFModelPredictionAnimationRaw(model, x, y, anim, colinfo(data, exclude), steps)

def TFModelPredictionAnimationRaw(model, x:str, y:str, anim:str, coldata:List[Dict], steps:int=20) -> pd.DataFrame:
    """
    Creates a dataset from a 2d prediction on a tensorflow model. Wrapper function for TFModelPredictionGridRaw()
    that automatically handles column info generation.

    Call from TFModelPredictionGrid to autogen params.

    Coldata should be formatted with keys 'name', 'min', 'max', 'mean'

    :param model: A tensorflow model
    :param model: A tensorflow model
    :param x: xaxis for graph data
    :param y: yaxis for graph data
    :param anim: Animation axis for graph data
    :param coldata: An ordered list of dicts with col names, min max values, and means
    :param steps: Resolution to scan model with
    """

    allcols = []
    for item in coldata:
        allcols.append(item['name'])

    assert x in allcols, "X must be in coldata"
    assert y in allcols, "Y must be in coldata"
    assert anim in allcols, "Anim must be in coldata"

    cols = []
    for item in coldata:
        if item not in [x, y, anim]:
            cols.append(item['name'])

    srow = []
    for item in cols:
        for d in coldata:
            if d['name'] == item:
                srow.append(d['mean'])

    srow = [srow] * (steps ** 3)
    preddata = pd.DataFrame(srow, columns=cols)

    col = []
    for pos in range(0, steps):
        for item in coldata:
            if item['name'] == x:
                col.append(pos * (item['max'] - item['min']) / (steps - 1) + item['min'])
    col = col * (steps ** 2)
    preddata[x] = col

    col = []
    for pos in range(0, steps):
        for item in coldata:
            if item['name'] == y:
                col += [pos * (item['max'] - item['min']) / (steps - 1) + item['min']] * steps
    col = col * steps
    preddata[y] = col

    col = []
    for pos in range(0, steps):
        for item in coldata:
            if item['name'] == anim:
                col += [pos * (item['max'] - item['min']) / (steps - 1) + item['min']] * (steps ** 2)
    preddata[anim] = col

    predictions = model.predict(preddata)
    preddata['Output'] = predictions
    return preddata

#endregion
#endregion



from MLVisualizationTools.modelanalytics import analyzeTFModel, analyzeTFModelRaw
from MLVisualizationTools.modelinterface import TFModelPredictionGrid, TFModelPredictionGridRaw
from MLVisualizationTools.modelinterface import TFModelPredictionAnimation, TFModelPredictionAnimationRaw
from MLVisualizationTools.graphinterface import plotlyGrid, plotlyAnimation, matplotlibGrid
from enum import Enum as enum

#Don't use these, __init__ is better

#A bunch of wrapper functions to make calling various tools easier

class Enum(enum):
    def __contains__(self, item):
        return item in [v.value for v in self.__members__.values()]

class ModelTypes(Enum):
    Tensorflow = "tf"

class InterfaceTypes(Enum):
    Grid = "grid"
    Animation = "animation"

class DataTypes(Enum):
    Grid = "grid"
    Animation = "animation"

class OutputTypes(Enum):
    Plotly = "Plotly"
    Matplotlib = "Matplotlib"

class Analytics:
    Tensorflow = analyzeTFModel
    TensorflowRaw = analyzeTFModelRaw

class Interfaces:
    TensorflowGrid = TFModelPredictionGrid
    TensorflowGridRaw = TFModelPredictionGridRaw
    TensorflowAnimation = TFModelPredictionAnimation
    TensorflowAnimationRaw = TFModelPredictionAnimationRaw

class Graphs:
    PlotlyGrid = plotlyGrid
    PlotlyAnimation = plotlyAnimation
    MatplotlibGrid = matplotlibGrid

class Mode:
    def __init__(self, modeltype = ModelTypes.Tensorflow, outputtype = OutputTypes.Plotly, useraw=False):
        if modeltype not in ModelTypes:
            raise Exception("Model type not found! (Try using the enum!)")
        if outputtype not in OutputTypes:
            raise Exception("Output type not found! (Try using the enum!)")
        self.modeltype = modeltype
        self.outputtype = outputtype
        self.useraw = useraw

    def analyzer(self):
        """Get an analyzer based on mode settings"""
        if self.modeltype == ModelTypes.Tensorflow:
            if self.useraw:
                return Analytics.TensorflowRaw
            else:
                return Analytics.Tensorflow

        else:
            raise Exception("No analytics found for this mode!")

    def interface(self, interface):
        """Get an ml model interface based on mode settings"""
        if interface not in InterfaceTypes:
            raise Exception("Interface type not found! (Try using the enum!)")

        if self.modeltype == ModelTypes.Tensorflow and interface == InterfaceTypes.Grid:
            if self.useraw:
                return Interfaces.TensorflowGridRaw
            else:
                return Interfaces.TensorflowGrid

        elif self.modeltype == ModelTypes.Tensorflow and interface == InterfaceTypes.Animation:
            if self.useraw:
                return Interfaces.TensorflowAnimationRaw
            else:
                return Interfaces.TensorflowAnimation
        else:
            raise Exception("No interface found for this mode!")

    def graph(self, datatype):
        """
        Get a graph based on mode settings
        """
        if datatype not in DataTypes:
            raise Exception("Data type not found! (Try using the enum!)")
        if self.outputtype == OutputTypes.Plotly and datatype == DataTypes.Grid:
            return Graphs.PlotlyGrid
        elif self.outputtype == OutputTypes.Plotly and datatype == DataTypes.Animation:
            return Graphs.PlotlyAnimation
        elif self.outputtype == OutputTypes.Matplotlib and datatype == DataTypes.Grid:
            return Graphs.MatplotlibGrid
        else:
            raise Exception("No graph found for this mode!")


from MLVisualizationTools.backend import colinfo
from typing import List, Dict
import copy
import pandas as pd

#Functions for retrieving data about ml model structure

#TODO - nonlinear

class AnalyticsColumnInfo:
    """Wrapper class for holding col info"""
    def __init__(self, name: str, variance: float):
        self.name = name
        self.variance = variance

    def __lt__(self, other):
        return self.variance < other.variance

    def __repr__(self):
        return "Col with name: " + self.name + " and variance " + str(self.variance)

class AnalyticsResult:
    """Wrapper class for holding and processing col info"""
    def __init__(self):
        self.cols: List[AnalyticsColumnInfo] = []

    def append(self, name: str, variance: float):
        self.cols.append(AnalyticsColumnInfo(name, variance))

    def maxVariance(self):
        """Return a list of cols, ordered by maximum variance"""
        cols = copy.copy(self.cols)
        cols.sort(reverse=True)
        return cols

#region Tensorflow
def analyzeTFModel(model, data: pd.DataFrame, exclude: List[str] = None, steps:int=20) -> AnalyticsResult:
    """
    Performs 1d analysis on a tensorflow model. Wrapper function for analyzeTFModelRaw()
    that automatically handles column info generation.

    :param model: A tensorflow model
    :param data: A pandas dataframe
    :param exclude: Values to be excluded from data, useful for output values
    :param steps: Resolution to scan model with
    """
    print("<BACKEND> Calling tfModelRaw")
    return analyzeTFModelRaw(model, colinfo(data, exclude), steps)

def analyzeTFModelRaw(model, coldata: List[Dict], steps:int=20) -> AnalyticsResult:
    """
    Performs 1d analysis on a tensorflow model. Returns a class with lots of info for graphing.
    Call from anaylyzeTFModel to autogen params.

    Coldata should be formatted with keys 'name', 'min', 'max', 'mean'

    :param model: A tensorflow model
    :param coldata: An ordered list of dicts with col names, min max values, and means
    :param steps: Resolution to scan model with
    """
    print("<BACKEND> Creating AR")
    AR = AnalyticsResult()
    print("<BACKEND> Prepping data structure")

    predrow = []
    cols = []
    for item in coldata:
        predrow.append(item['mean'])
        cols.append(item['name'])
    predrow = [predrow] * (steps * len(coldata))
    preddata = pd.DataFrame(predrow, columns=cols)

    currentpos = 0
    for item in coldata:
        for i in range(0, steps):
            preddata[item['name']][i + currentpos] = i * (item['max'] - item['min'])/(steps-1) + item['min']
        currentpos += steps

    print("<BACKEND> Calling model predict")
    print("<BACKEND> Model type: ", type(model))
    predictions = model.predict(preddata)
    print("<BACKEND> Model predict finished")

    currentpos = 0
    for item in coldata:
        values = predictions[currentpos:currentpos + steps]
        currentpos += steps
        AR.append(item['name'], values.max() - values.min())
    print("<BACKEND> AR finished")
    return AR
#endregion


from MLVisualizationTools import Analytics, Interfaces, Graphs, Colorizers
from MLVisualizationTools.backend import fileloader
import pandas as pd
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2' #stops agressive error message printing
from tensorflow import keras

#TODO - data overlay

try:
    import plotly
except:
    raise ImportError("Plotly is required to run this demo. If you don't have plotly installed, install it with"
                      " `pip install plotly' or run the matplotlib demo instead.")

def main():
    model = keras.models.load_model(fileloader('examples/Models/titanicmodel'))
    df: pd.DataFrame = pd.read_csv(fileloader('examples/Datasets/Titanic/train.csv'))

    AR = Analytics.Tensorflow(model, df, ["Survived"])
    maxvar = AR.maxVariance()

    grid = Interfaces.TensorflowGrid(model, maxvar[0].name, maxvar[1].name, df, ["Survived"])
    grid = Colorizers.Binary(grid)
    fig = Graphs.PlotlyGrid(grid, maxvar[0].name, maxvar[1].name)
    fig.show()

    grid = Interfaces.TensorflowGrid(model, 'Parch', 'SibSp', df, ["Survived"])
    grid = Colorizers.Binary(grid, highcontrast=True)
    fig = Graphs.PlotlyGrid(grid, 'Parch', 'SibSp')
    fig.show()

if __name__ == "__main__":
    main()


import pandas as pd

def simpleColor(data: pd.DataFrame, color) -> pd.DataFrame:
    """Marks all points as being the color inputted"""
    data['Color'] = [color] * len(data)
    return data

#TODO - keys and set default to highcontrast
def binaryColor(data: pd.DataFrame, highcontrast:bool=False, truecolor=None, falsecolor=None,
                cutoff:float=0.5, outputkey: str = 'Output') -> pd.DataFrame:
    """
    Colors grid based on whether the value is higher than the cutoff. Default colors are green for true and red
    for false. Black will appear if an error occurs.

    :param data: Input data
    :param highcontrast: Switches default colors to blue for true and orange for false
    :param truecolor: Manually specify truecolor
    :param falsecolor: Manually specify falsecolor
    :param cutoff: Cutoff value, higher is true
    :param outputkey: Key to grab values from
    """
    if truecolor is None:
        if not highcontrast:
            truecolor = "green"
        else:
            truecolor = "blue"
    if falsecolor is None:
        if not highcontrast:
            falsecolor = "red"
        else:
            falsecolor = "orange"

    #grid['Color'] = 'black' * len(grid)
    data.loc[data[outputkey] > cutoff, 'Color'] = truecolor
    data.loc[data[outputkey] <= cutoff, 'Color'] = falsecolor

    return data



"""
A set of functions and demos to make machine learning projects easier to understand through effective visualizations.
"""

from MLVisualizationTools.modelanalytics import analyzeTFModel, analyzeTFModelRaw
from MLVisualizationTools.modelinterface import TFModelPredictionGrid, TFModelPredictionGridRaw
from MLVisualizationTools.modelinterface import TFModelPredictionAnimation, TFModelPredictionAnimationRaw
from MLVisualizationTools.graphinterface import plotlyGrid, plotlyAnimation, matplotlibGrid
from MLVisualizationTools.colorizer import simpleColor, binaryColor

#A bunch of wrapper functions to make calling various tools easier
class Analytics:
    Tensorflow = analyzeTFModel
    TensorflowRaw = analyzeTFModelRaw

class Interfaces:
    TensorflowGrid = TFModelPredictionGrid
    TensorflowGridRaw = TFModelPredictionGridRaw
    TensorflowAnimation = TFModelPredictionAnimation
    TensorflowAnimationRaw = TFModelPredictionAnimationRaw

class Colorizers:
    Simple = simpleColor
    Binary = binaryColor

class Graphs:
    PlotlyGrid = plotlyGrid
    PlotlyAnimation = plotlyAnimation
    MatplotlibGrid = matplotlibGrid


from typing import List, Dict
import pandas as pd
from os import path

#Backend functions and classes used by the other scripts

def colinfo(data: pd.DataFrame, exclude:List[str] = None) -> List[Dict]:
    """
    Helper function for generating column info dict for a datframe

    :param data: A pandas Dataframe
    :param exclude: A list of data items to exclude
    """
    if exclude is None:
        exclude = []

    coldata = []
    for item in data.columns:
        if item not in exclude:
            coldata.append({'name': item, 'mean': data[item].mean(),
                            'min': data[item].min(), 'max': data[item].max()})
    return coldata

def fileloader(target: str):
    """Specify a path relative to MLVisualizationTools"""
    return path.dirname(__file__) + '/' + target


from MLVisualizationTools import Analytics, Interfaces, Graphs, Colorizers
from MLVisualizationTools.dashbackend import getTheme, getDashApp
import pandas as pd

#TODO - tour

try:
    import dash
    from dash import Input, Output
    from dash import dcc
    from dash import html
    import dash_bootstrap_components as dbc
    import plotly
except ImportError:
    raise ImportError("Dash and plotly are required to use this tool. Install them with the [dash] flag"
                      " on installation of this library.")

class App:
    def __init__(self, model, data: pd.DataFrame, title:str = "DashModelVisualizer", theme:str = "dark", folder = None,
                 highcontrast:bool = True, notebook:bool = False, kagglenotebook:bool = False, mode:str = 'external',
                 host:str = '0.0.0.0', port: bool = None):

        print("<DASH APP> Inside init function")
        theme, folder, self.figtemplate = getTheme(theme, folder)
        print("<DASH APP> Loaded theme")
        self.app, self.runFunc = getDashApp(title, notebook, kagglenotebook, host, port, mode, theme, folder)
        print("<DASH APP> Retrieved dash app")

        self.model = model
        self.df = data
        self.highcontrast = highcontrast

        options = []
        for col in self.df.columns:
            options.append({'label': col, 'value': col})

        print("<DASH APP> Before model processing")
        self.AR = Analytics.Tensorflow(self.model, self.df)
        print("<DASH APP> After model processing")
        self.maxvar = self.AR.maxVariance()

        self.x = self.maxvar[0].name
        self.y = self.maxvar[1].name

        self.fig = self.updateGraph()
        print("<DASH APP> Updated graph")

        graph = dbc.Card([
            dcc.Graph(id='example-graph', figure=self.fig)
        ], body=True)

        config = dbc.Card([
            dbc.Label("X Axis: "),
            dcc.Dropdown(id='xaxis', options=options, value=self.x),
            html.Br(),
            dbc.Label("Y Axis: "),
            dcc.Dropdown(id='yaxis', options=options, value=self.y),
            html.Br(),
        ], body=True)

        self.app.layout = dbc.Container([
            html.H1(title),
            html.Hr(),
            dbc.Row([
                dbc.Col(config, md=4),
                dbc.Col(graph, md=8)]
            ),
            html.P()],
            fluid=True,
            className='dash-bootstrap'
        )
        print("<DASH APP> Dash app structure")

        inputs = [Input('xaxis', "value"), Input('yaxis', 'value')]
        self.app.callback(Output("example-graph", "figure"), inputs)(self.updateGraphFromWebsite)
        print("<DASH APP> Registed callbacks")

    def run(self):
        print("Run function being called.")
        self.runFunc()
        print("Run function done, waiting on thread to stop process.")

    def updateGraph(self):
        data = Interfaces.TensorflowGrid(self.model, self.x, self.y, self.df)
        data = Colorizers.Binary(data, highcontrast=self.highcontrast)
        self.fig = Graphs.PlotlyGrid(data, self.x, self.y)
        self.fig.update_layout(template=self.figtemplate)
        return self.fig

    def updateGraphFromWebsite(self, x, y):
        self.x = x
        self.y = y
        return self.updateGraph()

def visualize(model, data: pd.DataFrame, title:str = "DashModelVisualizer", theme:str = "dark", folder = None,
              highcontrast:bool = True, notebook:bool = False, kagglenotebook:bool = False, mode:str = 'external',
              host:str = '0.0.0.0', port: bool = None):
    """
    Creates a dash website to visualize an ML model.

    :param model: A tensorflow keras model
    :param data: A pandas dataframe, all df columns must be numerical model inputs
    :param title: Title for website
    :param theme: Theme to load app in, can be a string (light / dark) or a url to load a stylesheet from
    :param folder: Directory to load additional css and js from
    :param highcontrast: Visualizes the model with orange and blue instead of green and red. Great for colorblind people!
    :param notebook: Uses jupyter dash instead of dash
    :param kagglenotebook: Enables ngrok tunneling for use in kaggle notebooks
    :param mode: Use 'external', 'inline', or 'jupyterlab'
    :param host: default hostname for dash
    :param port: None for default port (8050) or (1005)
    """
    App(model, data, title, theme, folder, highcontrast, notebook, kagglenotebook, mode, host, port).run()